import 'dart:io';

import 'package:semver_parser/src/semver.dart';

const _printUsageString = "usage";

/// Runs a CLI that reads from stdin and attempts to parse the input. Supports parsing input strings or various comparision operators.
///
/// To Run: 'dart example/semver_parsing_example.dart' or 'pub run example/semver_parsing_example.dart'.
///
/// This is intended as a demo or testing tool for the semver parsing in this library.
void main() {
  _printUsageOutput();

  for (;;) {
    stdout.writeln("Input?");

    final inputSemverString = stdin.readLineSync();

    _parseInput(inputSemverString);

    // add a blank line for spacing
    stdout.writeln("");
  }
}

final List<_ComparisonOperator> _supportedCompareOperators = [
  new _EqualityOperator(),
  new _NotEqualOperator()
];

void _printUsageOutput() {
  stdout..writeln("Input a string to parse it.\n\tE.X. '1.2.3'");

  for (var compareOp in _supportedCompareOperators) {
    stdout.writeln(compareOp.usageExample());
  }

  stdout.writeln("\nPrint this output again by inputting '$_printUsageString'\n");
}

bool _shouldPrintUsageOutput() {
  stdout.writeln("Print usage information and examples? (y or n)");

  return stdin.readLineSync() == "y";
}

void _parseInput(String inputString) {
  // If the input is the _printUsageString and the user confirms it, print usage information
  if (inputString == _printUsageString && _shouldPrintUsageOutput()) {
    _printUsageOutput();
    return;
  }

  // See if one (and only one) supported comparision operator matches the input and run it (assuming the user confirms it)
  final compareOp = _compareOperatorMatch(inputString);
  if (compareOp != null && compareOp.shouldRunCompare()) {
    compareOp.runCompare(inputString);
    return;
  }

  // Try to parse the input as a single Semantic Version
  _attemptSemverParse(inputString);
}

/// Determines if one (and only one) [_ComparisonOperator]s in [_supportedCompareOperators] match the input.
/// Returns the [_ComparisonOperator] that matched.
_ComparisonOperator _compareOperatorMatch(String input) {
  final compareMatches = _supportedCompareOperators.where((compare) => compare.matches(input));

  if (compareMatches.length != 1) {
    // too many or too few matches, do not try to handle the input
    return null;
  }

  return compareMatches.elementAt(0);
}

/// Attempts to parse the [inputString] as a Semantic Version.
/// Writes the result (successfully or exception) to [stdout].
Semver _attemptSemverParse(String inputString) {
  Semver semver;
  try {
    semver = parseSemverString(inputString);
    stdout.writeln("Parsed input as Semantic Version: '$semver'");
  } on SemverParseException catch (e) {
    stdout.writeln("Parsed input ('$inputString') threw: '$e");
  }

  return semver;
}

/// Defines the base for parsing an input string as comparing two Semantic Versions.
abstract class _ComparisonOperator {
  /// Return the string of the [_ComparisonOperator] (E.X. '==')
  String getOperatorString();

  /// Return a simple description and example for printing in the CLI usage output.
  String usageExample();

  /// Return if the [inputString] matches this [_ComparisonOperator].
  bool matches(String inputString);

  /// The function to run to compare two [Semver] objects.
  bool compareFunction(Semver one, Semver two);

  /// Return if the comparison should be run by asking the user.
  bool shouldRunCompare() {
    stdout.writeln(
        "Detected a comparison operator (${getOperatorString()}) in the input, attempt comparing semver strings? (y or n)");

    final answerString = stdin.readLineSync();

    return answerString == "y";
  }

  /// Parses out two [Semver] objects from the [inputString] and runs the given [compareFunction] against them.
  /// Will tell the user the result (or if parsing either [Semver] string fails).
  void runCompare(String inputString) {
    final versions = inputString.split(getOperatorString());

    final one = _attemptSemverParse(versions[0]);
    if (one == null) {
      stdout.writeln("First semver string ('${versions[0]}') is invalid");
      return;
    }

    final two = _attemptSemverParse(versions[1]);
    if (two == null) {
      stdout.writeln("Second semver string ('${versions[1]}') is invalid");
      return;
    }

    stdout.writeln("$one ${getOperatorString()} $two is: ${compareFunction(one, two)}");
  }
}

/// Defines the equality (==) operator for comparing Semantic Versions.
class _EqualityOperator extends _ComparisonOperator {
  @override
  String getOperatorString() => "==";

  @override
  String usageExample() => "Compare two versions for equality.\n\tE.x. '1.2.3==1.2.3'";

  @override
  bool compareFunction(Semver one, Semver two) => one == two;

  @override
  bool matches(String inputString) =>
      "=".allMatches(inputString).length == 2 && "==".allMatches(inputString).length == 1;
}

/// Defines the not equals (!=) operator for comparing Semantic Versions.
class _NotEqualOperator extends _ComparisonOperator {
  @override
  String getOperatorString() => "!=";

  @override
  String usageExample() => "See if two versions are not equal.\n\tE.x. '5.12.1!=5.13.2'";

  @override
  bool compareFunction(Semver one, Semver two) => one != two;

  @override
  bool matches(String inputString) =>
      "=".allMatches(inputString).length == 1 &&
      "!".allMatches(inputString).length == 1 &&
      "!=".allMatches(inputString).length == 1;
}
