/// [Semver] defines a simple class to hold information about a single Semantic Version.
///
/// This would be major version, minor version and patch version.
/// E.X. '2.5.4' 2 = major, 5 = minor, 4 = patch.
/// Semver objects support comparision via the standard == and != operators.
class Semver {
  /// The major version of this Semantic Version.
  final int majorVersion;

  /// The minor version of this Semantic Version.
  final int minorVersion;

  /// The patch version of this Semantic Version.
  final int patchVersion;

  /// Constructs a [Semver] object from given major, minor and patch [int]s
  Semver(this.majorVersion, this.minorVersion, this.patchVersion);

  @override
  bool operator ==(Object other) {
    if (other is Semver) {
      return majorVersion == other.majorVersion &&
          minorVersion == other.minorVersion &&
          patchVersion == other.patchVersion;
    }

    return false;
  }

  @override
  int get hashCode => toString().hashCode;

  @override
  String toString() => "$majorVersion.$minorVersion.$patchVersion";
}

// This regex can parse a string and produce capture groups for major, minor and patch versions.
final RegExp _semverParseRegex = new RegExp(r"([0-9]+?)\.([0-9]+?)\.([0-9]+)");

/// Takes in a string and attempts to extract a valid Semantic Version from it.
Semver parseSemverString(String semverString) {
  final matches = _semverParseRegex.allMatches(semverString);

  if (matches.length != 1) {
    if (matches.isEmpty) {
      throw new SemverParseException.noMatches();
    }

    throw new SemverParseException.multipleMatches();
  }

  final match = matches.elementAt(0);
  final major = _parseSingleVersionString(match[1], () => new SemverParseException.invalidMajor());
  final minor = _parseSingleVersionString(match[2], () => new SemverParseException.invalidMinor());
  final patch = _parseSingleVersionString(match[3], () => new SemverParseException.invalidPatch());

  return new Semver(major, minor, patch);
}

/// Parses a string into an int. If the parse fails or the string starts with a '0' (which would be invalid
/// in a version) then [exceptionCallback] will be called and the [SemverParseException] returned thrown.
int _parseSingleVersionString(String versionString, SemverParseException exceptionCallback()) {
  // a version starting with a leading 0 is invalid (unless it is only '0')
  if (versionString.length > 1 && versionString[0] == "0") {
    throw exceptionCallback();
  }

  return int.parse(versionString, onError: (_) => throw exceptionCallback());
}

/// An exception generated when a string cannot be parsed to a single Semantic Version.
class SemverParseException implements Exception {
  /// Exception message if the string has no Semantic Version in it.
  static const String noMatchesMessage = "No matches";

  /// Exception message if the string has multiple Semantic Versions in it.
  static const String multipleMatchesMessage = "Multiple matches";

  /// Exception message if the string match for the major version is invalid.
  static const String invalidMajorVersionMessage = "Major version is invalid";

  /// Exception message if the string match for the minor version is invalid.
  static const String invalidMinorVersionMessage = "Minor version is invalid";

  /// Exception message if the string match for the patch version is invalid.
  static const String invalidPatchVersionMessage = "Patch version is invalid";

  /// Human readable message describing why the string couldn't be parsed to a Semantic Version.
  final String failureMessage;

  /// Create a [SemverParseException] for no matches.
  SemverParseException.noMatches() : this.failureMessage = noMatchesMessage;

  /// Create a [SemverParseException] for multiple matches.
  SemverParseException.multipleMatches() : this.failureMessage = multipleMatchesMessage;

  /// Create a [SemverParseException] for invalid major version.
  SemverParseException.invalidMajor() : this.failureMessage = invalidMajorVersionMessage;

  /// Create a [SemverParseException] for invalid minor version.
  SemverParseException.invalidMinor() : this.failureMessage = invalidMinorVersionMessage;

  /// Create a [SemverParseException] for invalid patch version.
  SemverParseException.invalidPatch() : this.failureMessage = invalidPatchVersionMessage;

  @override
  String toString() => "SemverParseException message: '$failureMessage'";
}
