import "package:test/test.dart";

import 'package:semver_parser/src/semver.dart';

final _validTestsCase = [
  // cases where the string is nothing but a semantic version (major.minor.patch)
  new _SemverValidTestCase("simple", "1.2.3", 1, 2, 3),
  new _SemverValidTestCase("zero major", "0.5.6", 0, 5, 6),
  new _SemverValidTestCase("zero minor", "4.0.6", 4, 0, 6),
  new _SemverValidTestCase("zero patch", "7.12.0", 7, 12, 0),
  new _SemverValidTestCase("double digit major", "13.2.5", 13, 2, 5),
  new _SemverValidTestCase("double digit minor", "7.22.0", 7, 22, 0),
  new _SemverValidTestCase("double digit patch", "4.5.11", 4, 5, 11),
  new _SemverValidTestCase("triple digit major", "156.45.6", 156, 45, 6),
  new _SemverValidTestCase("triple digit minor", "1.457.2", 1, 457, 2),
  new _SemverValidTestCase("triple digit patch", "6.22.145", 6, 22, 145),
  // cases where the string contains a semantic version (major.minor.patch),
  // but also has some "junk" starting and/or ending text
  new _SemverValidTestCase("prefix text", "package12.3.52", 12, 3, 52),
  new _SemverValidTestCase("suffix text", "4.52.143cat4", 4, 52, 143),
  new _SemverValidTestCase("prefix and suffix text", "stuff9.3.64end", 9, 3, 64),
  new _SemverValidTestCase("prefix and suffix are number words", "nine3.17.4four", 3, 17, 4),
  new _SemverValidTestCase("prefix contains partial version", "dog1.4.cat12.4.5", 12, 4, 5),
  new _SemverValidTestCase("suffix contains partial version", "6.11.4hi5.6bye", 6, 11, 4),
  new _SemverValidTestCase(
      "prefix and suffix contains partial version", "chi4.cken6.7.11pup6.py", 6, 7, 11),
];

final _invalidTestCases = [
  // no numbers
  new _SemverInvalidTestCase("Words", "I am Dog", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Semver pattern, but no numbers", "x.y.z", SemverParseException.noMatchesMessage),

  // not enough numbers
  new _SemverInvalidTestCase("Only two numbers", "4.5", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Only two numbers, leading period", ".6.13", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Only two numbers, ending period", "2.9.", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase("Only two numbers, leading and ending period", ".0.2.",
      SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase("Only one number", "3", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Only one number, leading period", ".6", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Only one number, ending period", "12.", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Only one number, leading and ending period", ".7.", SemverParseException.noMatchesMessage),

  // too many numbers
  new _SemverInvalidTestCase(
      "Six version semantic version", "2.3.4.5.6.7", SemverParseException.multipleMatchesMessage),
  new _SemverInvalidTestCase("Seven version semantic version", "12.5.123.7.33.5.112",
      SemverParseException.multipleMatchesMessage),

  // leading zeros are invalid
  new _SemverInvalidTestCase(
      "Leading zero on major version", "012.4.11", SemverParseException.invalidMajorVersionMessage),
  new _SemverInvalidTestCase(
      "Leading zero on minor version", "7.0124.1", SemverParseException.invalidMinorVersionMessage),
  new _SemverInvalidTestCase(
      "Leading zero on patch version", "4.15.02", SemverParseException.invalidPatchVersionMessage),
  new _SemverInvalidTestCase("Multiple lseading zeros on major version", "000012.4.11",
      SemverParseException.invalidMajorVersionMessage),
  new _SemverInvalidTestCase("LMultiple lseading zeros on minor version", "7.00124.1",
      SemverParseException.invalidMinorVersionMessage),
  new _SemverInvalidTestCase("Multiple lseading zeros on patch version", "4.15.0000000002",
      SemverParseException.invalidPatchVersionMessage),

  // one valid version, two invalid ones
  new _SemverInvalidTestCase(
      "Major is numberic, others are not", "5.t.y", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Minor is numberic, others are not", "h.67.y", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Patch is numberic, others are not", "o.t.934", SemverParseException.noMatchesMessage),

  // two valid versions, one invalid ones
  new _SemverInvalidTestCase(
      "All but major are valid", "n.15.2", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "All but minor are valid", "3.b.2", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "All but patch are valid", "12.13.p", SemverParseException.noMatchesMessage),

  // non-numeric characters mixed with otherwise valid patterns
  new _SemverInvalidTestCase(
      "Major has non-numeric character", "3t.7.4", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Minor has non-numeric character", "7.5-.19", SemverParseException.noMatchesMessage),
  new _SemverInvalidTestCase(
      "Patch has non-numeric character", "4.15.%43", SemverParseException.noMatchesMessage),

  // "one off" cases
  new _SemverInvalidTestCase(
      "Commas rather than periods", "3,4,8", SemverParseException.noMatchesMessage),

  // No test cases for failing to parse major, minor or patch version string since the regex in the parser
  // will only match valid and non-negative number strings and the dart vm doesn't have int overflow failures.
  // Well, technically it does if the int uses more RAM than available, but let's not try testing that ok? :)
];

final _equalityTestCases = [
  // two Semver objects should be equal
  new _EqualityTest("Basic equals", new Semver(1, 2, 3), new Semver(1, 2, 3)),
  new _EqualityTest("Large versions", new Semver(12, 43, 125), new Semver(12, 43, 125)),
  new _EqualityTest("Huge versions", new Semver(12654, 534523, 3523654748352),
      new Semver(12654, 534523, 3523654748352)),
];

final _notEqualTestCase = [
  // Semver objects have one different version
  new _NotEqualTest("Major version differs", new Semver(1, 2, 3), new Semver(2, 2, 3)),
  new _NotEqualTest("Minor version differs", new Semver(1, 2, 3), new Semver(1, 7, 3)),
  new _NotEqualTest("Patch version differs", new Semver(1, 2, 3), new Semver(1, 2, 11)),

  // Semver objects have two different versions
  new _NotEqualTest("Major and minor version differs", new Semver(1, 2, 3), new Semver(2, 13, 3)),
  new _NotEqualTest("Minor and patch version differs", new Semver(1, 2, 3), new Semver(1, 2, 1)),
  new _NotEqualTest("Minor and patch version differs", new Semver(1, 2, 3), new Semver(1, 8, 7)),

  // Semver objects have three different versions
  new _NotEqualTest("All versions differ", new Semver(9, 8, 7), new Semver(1, 2, 3)),

  // not equal corner cases
  new _NotEqualTest("Compare to null", new Semver(1, 2, 3), null),
  new _NotEqualTest("Compare to non-Semver object", new Semver(1, 2, 3), new DateTime.now()),
  // the underlying hashCode implementation matches the string hashCode, this is ok as long as equality is still false
  new _NotEqualTest("Compare to matching semver string", new Semver(1, 2, 3), "1.2.3",
      ignoreHashCheck: true),
];

void main() {
  group('parseSemverString -', () {
    group('valid semver strings -', () {
      for (var testCase in _validTestsCase) {
        test(testCase.caseName, testCase.runTestCase);
      }
    });

    group('invalid semver strings -', () {
      for (var testCase in _invalidTestCases) {
        test(testCase.caseName, testCase.runTestCase);
      }
    });

    group('semver equality -', () {
      for (var testCase in _equalityTestCases) {
        test(testCase.caseName, testCase.runTestCase);
      }

      test('Same object equality', () {
        final semver = new Semver(4, 5, 12);

        // A==A compare
        expect(semver == semver, isTrue, reason: "Semver object should equal itself");
        expect(semver.hashCode == semver.hashCode, isTrue,
            reason: "Semver object hashCode should equal itself");

        // A!=A compare
        expect(semver != semver, isFalse, reason: "Semver object should not be != to itself");
        expect(semver.hashCode != semver.hashCode, isFalse,
            reason: "Semver object hashCode should not be != to itself");
      });
    });

    group('semver not equal -', () {
      for (var testCase in _notEqualTestCase) {
        test(testCase.caseName, testCase.runTestCase);
      }
    });
  });
}

class _SemverValidTestCase {
  final String caseName;

  final String testString;

  final int expectedMajor;
  final int expectedMinor;
  final int expectedPatch;

  _SemverValidTestCase(
      this.caseName, this.testString, this.expectedMajor, this.expectedMinor, this.expectedPatch);

  void runTestCase() {
    Semver semver;

    try {
      semver = parseSemverString(testString);
    } on SemverParseException catch (e) {
      fail('Parsing semver string failed with SemverParseException $e');
    }

    expect(semver, isNotNull, reason: "semver object shoudn't be null");
    expect(semver.majorVersion, equals(expectedMajor), reason: "Major version is wrong");
    expect(semver.minorVersion, equals(expectedMinor), reason: "Minor version is wrong");
    expect(semver.patchVersion, equals(expectedPatch), reason: "Patch version is wrong");

    final expectedString = "$expectedMajor.$expectedMinor.$expectedPatch";
    expect("$semver", equals(expectedString), reason: "Semver ToString is wrong");
  }
}

class _SemverInvalidTestCase {
  final String caseName;

  final String testString;

  final String expectedExceptionMessage;

  _SemverInvalidTestCase(this.caseName, this.testString, this.expectedExceptionMessage);

  void runTestCase() {
    SemverParseException exception;
    try {
      parseSemverString(testString);
    } on SemverParseException catch (e) {
      exception = e;
    }

    expect(exception, isNotNull,
        reason: "Invalid test case should have generated a SemverParseException");
    expect(exception.failureMessage, equals(expectedExceptionMessage),
        reason: "Incorrect exception for invalid case");
    expect(exception.toString(),
        equals("SemverParseException message: '${exception.failureMessage}'"));
  }
}

class _EqualityTest {
  final String caseName;

  // both are types as Object to allow testing equality against non-Semver objects
  final Object semverOne;
  final Object semverTwo;

  final bool ignoreHashCheck;

  _EqualityTest(this.caseName, this.semverOne, this.semverTwo, {this.ignoreHashCheck: false});

  void runTestCase() {
    _expectEqual();
    if (!ignoreHashCheck) {
      _expectHashEqual();
    }
  }

  void _expectEqual() {
    // A==B compare
    expect(semverOne == semverTwo, isTrue, reason: "SemverOne == SemverTwo should be true");

    // B==A compare
    expect(semverTwo == semverOne, isTrue,
        reason: "SemverTwo == SemverOne should be true, inverse compare");

    // A!=B compare
    expect(semverOne != semverTwo, isFalse, reason: "SemverOne != SemverTwo should be false");

    // B!=A compare
    expect(semverTwo != semverOne, isFalse,
        reason: "semverTwo != SemverOne should be false, inverse compare");
  }

  void _expectHashEqual() {
    // A==B compare
    expect(semverOne.hashCode == semverTwo.hashCode, isTrue,
        reason: "SemverOne.hasCode == SemverTwo.hashCode should be true");

    // B==A compare
    expect(semverTwo.hashCode == semverOne.hashCode, isTrue,
        reason: "SemverTwo.hasCode == SemverOne.hashCode should be true, inverse compare");

    // A!=B compare
    expect(semverOne.hashCode != semverTwo.hashCode, isFalse,
        reason: "SemverOne.hasCode != SemverTwo.hashCode should be false");

    // B!=A compare
    expect(semverTwo.hashCode != semverOne.hashCode, isFalse,
        reason: "semverTwo.hasCode != SemverOne.hashCode should be false, inverse compare");
  }
}

class _NotEqualTest {
  final String caseName;

  // both are types as Object to allow testing equality against non-Semver objects
  final Object semverOne;
  final Object semverTwo;

  final bool ignoreHashCheck;

  _NotEqualTest(this.caseName, this.semverOne, this.semverTwo, {this.ignoreHashCheck: false});

  void runTestCase() {
    _expectNotEqual();
    if (!ignoreHashCheck) {
      _expectHashNotEqual();
    }
  }

  void _expectNotEqual() {
    // A!=B compare
    expect(semverOne != semverTwo, isTrue, reason: "SemverOne != SemverTwo should be true");

    // B!=A compare
    expect(semverTwo != semverOne, isTrue,
        reason: "SemverTwo != SemverOne should be true, inverse compare");

    // A==B compare
    expect(semverOne == semverTwo, isFalse, reason: "SemverOne == SemverTwo should be false");

    // B==A compare
    expect(semverTwo == semverOne, isFalse,
        reason: "SemverTwo == SemverOne should be false, inverse compare");
  }

  void _expectHashNotEqual() {
    // A!=B compare
    expect(semverOne.hashCode != semverTwo.hashCode, isTrue,
        reason: "SemverOne.hasCode != SemverTwo.hashCode should be true");

    // B!=A compare
    expect(semverTwo.hashCode != semverOne.hashCode, isTrue,
        reason: "SemverTwo.hasCode != SemverOne.hashCode should be true, inverse compare");

    // A==B compare
    expect(semverOne.hashCode == semverTwo.hashCode, isFalse,
        reason: "SemverOne.hasCode == SemverTwo.hashCode should be false");

    // B==A compare
    expect(semverTwo.hashCode == semverOne.hashCode, isFalse,
        reason: "SemverTwo.hasCode == SemverOne.hashCode should be false, inverse compare");
  }
}
